// Version 2: 
//   Data is allocated on heap, with pointers to it on the stack. 
//   AMPI virtualization ratios > 1 are supported.
//   AMPI migration is not supported (no pup).

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>

#define BLOCKSIZE 2000

class chunk {
public:
  double t[BLOCKSIZE+2];
  int xm, xp;
};

#define abs(x) ((x)<0.0 ? -(x) : (x))

int main(int ac, char** av)
{
  int i, j, iter, niter;
  double error, tval, maxerr, tmpmaxerr;
  chunk *cp;
  int thisIndex, nblocks;
  double starttime, endtime, itertime;

  MPI_Init(&ac, &av);
  MPI_Comm_rank(MPI_COMM_WORLD, &thisIndex);
  MPI_Comm_size(MPI_COMM_WORLD, &nblocks);

  if (ac >= 2)
    niter = atoi(av[1]);
  else
    niter = 20;

  if (thisIndex == 0)
    printf("Running on %d MPI processors for %d iterations. \n", nblocks, niter);

  MPI_Bcast(&niter, 1, MPI_INT, 0, MPI_COMM_WORLD);

  // Data is allocated on the heap, with each AMPI rank performing its own allocation
#if CMK_AIX
  // On AIX, new is not intercepted by AMPI's malloc, so we must use malloc here
  cp = (chunk*)malloc(sizeof(chunk));
#else
  cp = new chunk;
#endif

  cp->xp = (thisIndex+1)%nblocks;
  cp->xm = (thisIndex+nblocks-1)%nblocks;
  for(i=1; i<=BLOCKSIZE; i++)
    cp->t[i] = thisIndex*BLOCKSIZE+i-1;

  MPI_Barrier(MPI_COMM_WORLD);

  for(iter=1; iter<=niter; iter++) {
    starttime = MPI_Wtime();
    maxerr = 0.0;

    MPI_Request rreq[2];
    MPI_Status rsts[2];

    MPI_Irecv(&cp->t[0], 1, MPI_DOUBLE, cp->xm, 1, MPI_COMM_WORLD, &rreq[0]);
    MPI_Irecv(&cp->t[BLOCKSIZE+1], 1, MPI_DOUBLE, cp->xp, 0,
	      MPI_COMM_WORLD, &rreq[1]);

    MPI_Send(&cp->t[1], 1, MPI_DOUBLE, cp->xm, 0, MPI_COMM_WORLD);
    MPI_Send(&cp->t[BLOCKSIZE], 1, MPI_DOUBLE, cp->xp, 1, MPI_COMM_WORLD);

    MPI_Waitall(2, rreq, rsts);

    for (j=0; j<thisIndex*2+5; j++)            // some load imbalance
    for(i=1; i<=BLOCKSIZE; i++) {
      tval = (cp->t[i] + cp->t[i-1] + cp->t[i+1] ) / 3.0;
      error = abs(tval-cp->t[i]);
      cp->t[i] = tval;
      if (error > maxerr) maxerr = error;
    }
    MPI_Allreduce(&maxerr, &tmpmaxerr, 1, MPI_DOUBLE, MPI_MAX,
		  MPI_COMM_WORLD);
    maxerr = tmpmaxerr;

    endtime = MPI_Wtime();
    itertime = endtime - starttime;
    double  it;
    MPI_Allreduce(&itertime, &it, 1, MPI_DOUBLE, MPI_SUM,
                   MPI_COMM_WORLD);
    itertime = it/nblocks;

    if (thisIndex == 0) printf("iter %d time: %lf maxerr: %lf\n", iter, itertime, maxerr);
  }
  MPI_Finalize();
  return 0;
}
